/*
 * Copyright (c) 2003-2009 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <stdlib.h>

#include <GPSNovatel/serial.h>
#include <GPSNovatel/msgParser.h>
#include <GPSNovatel/horodatage.h>
#include <GPSNovatel/logData.h>

// DEBUG
//#define DEBUG_ULONG
//#define DEBUG_USHORT
//#define DEBUG_HEADER
//#define DEBUG_DOUBLE
//#define DEBUG_FLOAT
//#define DEBUG_DETRAME
//#define DEBUG_OEM4_TRAME_COMPUTED

// Stockage des donnes dans un buffer
unsigned char GPS_TRAME[TAILLE_BUFFER];
unsigned char GPS_TRAME_COMPLETED[TAILLE_BUFFER];

// Variable de positionnement dans le buffer
// Indice qui donne la position courante pour la fin du buffer
// qui contient les donnes stockes
int LAST_POSITION=0;

// Variable qui donne la position dans le tableau
// des dates d'arrive des trames qui contiennent 0xAA
int LAST_POSITION_TIME=0;

// Buffer de lecture du port srie
unsigned char BUFFER_SERIE[TAILLE_BUFFER_SERIE];


struct timeval timeArrival[TAILLE_BUFFER];
struct timeval timeArrival_TMP[TAILLE_BUFFER];

/*!
  @brief Ouverture du port serie, renvoie -1 en cas de problme ou un
  file descriptor si ouverture russie. Cette fonction restitue l'tat du
  port srie aprs ouverture. L'ouverture est configurable selon VMIN et VTIME.
  L'utilisation des paramtres CARA et DELAIS respectivement VMIN, VTIME est dlicate
  et peut provoquer un mode blocant sur la lecture du port srie. Lire la documentation
  du projet sur l'utilisation de ces paramtres.

 @param const char *device : fichier systme qui reprsente le port srie (/dev/ttyS0 ...)
 @param int CARA : reprsente VMIN nombre de caractre
 @param int DELAIS : reprsente VTIME attente
 @param int vitesse : dbit choisi pour l'ouverture du port srie de 50  115200 bauds
*/

int openSerial(const char *device, int CARA, int DELAIS, int vitesse)
{
  int fd, flags;
  struct termios t;

  printf("com  : %s , VMIN = %d, VTIME = %d, speed = %i\n", device, CARA, DELAIS, vitesse);

  // on copie des 0 dans le buffer
  memset(GPS_TRAME, 0, TAILLE_BUFFER);

  // Ouverture du port srie en mode non bloquant, pour ne pas attendre la
  // porteuse
  fd = open(device, O_RDWR|O_NONBLOCK);
  if (fd < 0) {
    fprintf(stderr, "open : %s:%d: %s: %s\n", __FILE__, __LINE__,
	    device, strerror(errno));
    return -1;
  }

  // Initialisation d'une structure avec les paramtres
  // actuel du port srie
  tcgetattr(fd, &t);

  // Sauvegarde de la structure
  // TODO

  // Passage en mode raw -> on utilise pas le port comme un terminal
  // les caractres ne sont pas interprts mais juste lus
  cfmakeraw(&t);

  // On passe dans la vitesse demande
  if(vitesse == 50) cfsetspeed(&t, B50);
  else if(vitesse == 75) cfsetspeed(&t, B75);
  else if(vitesse == 110) cfsetspeed(&t, B110);
  else if(vitesse == 134) cfsetspeed(&t, B134);
  else if(vitesse == 150) cfsetspeed(&t, B150);
  else if(vitesse == 200) cfsetspeed(&t, B200);
  else if(vitesse == 300) cfsetspeed(&t, B300);
  else if(vitesse == 600) cfsetspeed(&t, B600);
  else if(vitesse == 1200) cfsetspeed(&t, B1200);
  else if(vitesse == 1800) cfsetspeed(&t, B1800);
  else if(vitesse == 2400) cfsetspeed(&t, B2400);
  else if(vitesse == 4800) cfsetspeed(&t, B4800);
  else if(vitesse == 9600) cfsetspeed(&t, B9600);
  else if(vitesse == 19200) cfsetspeed(&t, B19200);
  else if(vitesse == 38400) cfsetspeed(&t, B38400);
  else if(vitesse == 57600) cfsetspeed(&t, B57600);
  else if(vitesse == 115200) cfsetspeed(&t, B115200);

  // VMIN, VTIME
  t.c_cc[VMIN]= CARA;
  t.c_cc[VTIME]= DELAIS;

  // On applique les nouveaux attributs et on vide les buffers d'entre et de sortie
  // du port srie avec TCSAFLUSH
  if (tcsetattr(fd, TCSANOW, &t) < 0)
    {
      fprintf(stderr, "%s:%d: tcsetattr: %s\n", __FILE__, __LINE__,
	      strerror(errno));
      close (fd);
      return -1;
    }

  flags = fcntl(fd, F_GETFL, 0);

  if (flags < 0) {
    fprintf(stderr, "%s:%d: F_GETFL: %s\n", __FILE__, __LINE__,
	    strerror(errno));
    close (fd);
    return -1;
  }

  // On repasse en mode bloquant
/*  flags &= ~O_NONBLOCK;
  if (fcntl(fd, F_SETFL, flags) < 0) {
    fprintf(stderr, "%s:%d: F_SETFL: %s\n", __FILE__, __LINE__,
	    strerror(errno));
    close (fd);
    return -1;
  }
*/

  return fd;
}

/*!
  @brief Lecture des donnes brutes du port srie du GPS. On renvoie
  une stucture de donnes qui contient les informations du GPS pour les informations binaires
  et (TODO) si l'information est ascii, on le met dans un buffer (le retour des commands).

 @param int fd : descripteur de fichier du port srie
*/
int readSerial(int fd, int *nbTrame)
{
  ssize_t n;
  errno = 0;

  *nbTrame=-1;

  do
    {
      n = read(fd,BUFFER_SERIE, TAILLE_BUFFER_SERIE);
    }
  while(n < 0 && errno == EINTR);

  if (n < 0)
     {
	if (errno == EAGAIN)
	{
		*nbTrame = 0;
		return 0;
	} else
	{
		fprintf(stderr, "----> %s:%d: read: %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}
    }
  else
    {
      if(LAST_POSITION + n > TAILLE_BUFFER)
	{
	  // On peut considrer que si le buffer atteind une taille
	  // sup  la taille du buffer , c'est que les donnes n'ont pas t
	  // dpiles car les infos attendues ne sont pas conformes
	  // au protocole du GPS, on procde alors au vidage complet du buffer

	  // On nettoie le buffer de stockage
	  memset(GPS_TRAME, 0, TAILLE_BUFFER);

	  // On positionne le curseur  zro pour crire au dbut du buffer
	  LAST_POSITION = 0;

	  // Toutes les dates d'arrive sont remises  zro
	  memset(timeArrival_TMP, 0, TAILLE_BUFFER);
	}
      else if(n > 0) // on stock les data si on en recoit
	{
	  stockAsciiLog(BUFFER_SERIE,n);
	  stockBinaryTrame(BUFFER_SERIE, n);
	}

      // si la taille des donnes stockes est sup.  au moins
      // la taille d'un header alors on dtrame les donnes
      // dans le cas ou la trame commence  l'indice 0 du tableau
      // HEADER = 25 octets + 4 octets messages au Min + 4 octets CRC
      if(LAST_POSITION  > (25 + 4 + 4))
	{
	   *nbTrame = isTrameIsBuffered();
	}
    }

  return 0;
}


/*!
  @brief Stockage des donnes du GPS dans un buffer avec un curseur qui suit
  la dernre position on l'on a crit des donnes

  @param unsigned char *buf : le buffer lu du GPS
  @param int len : la taille des donnes lues
*/
void stockBinaryTrame(unsigned char *buf,  int len)
{
  int i;

  if(LAST_POSITION + len <  TAILLE_BUFFER)
    {

//printf("**************************** start RECU\n");

      // On copie le buffer recu  la suite du precedent buffer
      for(i=0;i<len;i++)
	{
	  // On copie le buffer lu dans le tableau qui contient
	  // toutes les trames
	  GPS_TRAME[LAST_POSITION] = buf[i];

	  // Si on trouve 0xAA dans le buffer, c'est peut tre un
	  // dbut de trame, on stocke donc une heure d'arrive, il faut
	  // ensuite faire la diffrence lors du traitement du buffer  diffrencier
	  // les octets == 0xAA et les vraies trames qui commencent aussi par 0xAA

	  //printf("%0x - ",buf[i]);

	  if (buf[i] == 0xaa)
	    {
	       stockArrivalTime(LAST_POSITION);
	    }

	  LAST_POSITION++;
	}
//printf("\n**************************** fin RECU\n");
    }
  // printf("LAST = %i\n", LAST_POSITION);
}


/*!
  TODO
 */
void stockAsciiLog(unsigned char *buf, int len) {
  (void) buf;
  (void) len;

}

/*!
  @brief On place dans un buffer la date d'arrive d'une ventuelle trame
  du GPS.

  @param int index
*/
void stockArrivalTime(int index)
{
  struct timeval timev;

  // Recuperation de l'heure systme
  gettimeofday(&timev, NULL);

//  printHorodatage(&timev);
//  printf("\nindex = %d\n", index);

  // Stockage du temps dans un tableau au mme indice
  // que le tableau ou sont stocks les donnes du GPS
  if(index <= TAILLE_BUFFER)
    {
      timeArrival[index].tv_sec = timev.tv_sec;

      timeArrival[index].tv_usec = timev.tv_usec;
    }
  else
    // si la taille est dpasse cela peut signifier que les dates
    // n'ont pas t dpiles car un problme  pu se poser
    {
       memset(timeArrival_TMP, 0, TAILLE_BUFFER);
    }
}


/*!
  @brief Cette fonction dpile le buffer dans lequel est stock les dates
  d'arrive des trames dans le systme. Cette date n'est pas la date exacte
  d'mission de la trame mais la date  laquelle le programme y accde. La date
  exacte est place dans le header binaire d'une trame. Cette procdure permet de
  dpiler les info de la position  bufferCompleted jusqu' lastPosition et de placer
  les infos dans un le buffer remis  zro.

 @param int bufferCompleted : position jusqu'a laquelle toutes les donnes du buffer
 ont t traites,
 @param int lastPosition : dernire position dans le buffer
*/
void unStackArrivalTime(int bufferCompleted,  int lastPosition)
{
  // On nettoie le buffer de stockage temporaire
  memset(timeArrival_TMP, 0, TAILLE_BUFFER);

  int i, ind=0;

  // On se place  la fin des heures traites dans le buffer et on rcupre
  // dans le buffer de sauvegarde les data qui ne sont pas encore traites
  for(i= bufferCompleted; i< lastPosition;i++)
    {
      timeArrival_TMP[ind] = timeArrival[i];

      ind++;
    }

  // On nettoie le buffer de stockage
  memset(timeArrival, 0, TAILLE_BUFFER);

  // On recopie dans le buffer de trame les infos qui n'ont pas t encore
  // traites, la distance est donne par (lastPosition - bufferCompleted)
  for(i= 0; i< lastPosition - bufferCompleted;i++)
    {
      timeArrival[i] = timeArrival_TMP[i];
    }
}

/*!
  Clear the serial port and buffer data. 
*/
int clearBufferSerie(int fd) {
  ssize_t n=0;
  errno = 0; 
  /* read persistant data */
  do {
    read(fd,BUFFER_SERIE,TAILLE_BUFFER_SERIE);
  } while (n < 0 && errno == EINTR);
  /* clear the buffer data structure */
  memset(BUFFER_SERIE, 0, TAILLE_BUFFER_SERIE);  
  return 0;
}

/*!
  @brief Cette procdure renvoie un tableau de structure dans lequel on
  retrouve les infos qui ont t recu par le GPS. On lit le buffer et on
  dtecte les arrive de trame avec les 3 octets de synchronisation. Puis
  si la trame est complte et si elle correspond  un type attendu alors on
  detrame le buffer. Cette opration renouvelle tant que possible. Les data
  sont stocke dans une structure puis on depile le buffer.
*/
int isTrameIsBuffered()
{
  int i;
  unsigned char  HEADER_LENGTH;
  unsigned short MESSAGE_LENGTH;

  int BUFFER_TRAITE=0;

  // le nombre de trame dtrame est remis  zro  chaque fois que l'on
  // traite le buffer de rception
  NB_TRAME_BUFFERED = 0;

  // initialisation de la structure qui stocke les donnes dtrames
  memset(USR_TRAME, 0, sizeof(OEM4_TRAME_COMPUTED));

#ifdef DEBUG_OEM4_TRAME_COMPUTED
  printf("\n\n############################################## DEBUT BUFFER AVANT TRAITEMENT\n");
  int debut;

  for(debut=0;debut<LAST_POSITION;debut++)
    {
      printf("%0x - ",GPS_TRAME[debut]);
    }
  printf("\n\n############################################## FIN BUFFER AVANT TRAITEMENT\n\n");
#endif


  // printf("\n");

  // on parcours le tableau qui stocke les trames en commencant
  //  la position prcdente
  for(i=0;i<LAST_POSITION;i++)
    {

    //   printf("%0x +", GPS_TRAME[i]);

      // On parcours le buffer pour trouver une trame au format binaire
      // Distance entre 0XAA et fin buffer > 25 octets + 4 octets de message Min + 4 octets CRC ?
      // pour que l'on puisse au moins avoir 1 header avec un message
      if (GPS_TRAME[i] == 0xaa && LAST_POSITION-i>= (25 + 4 + 4))
	{
	  // printf("------> DEBUT, i = %d\n", i);

	  // On compte le nb de trame ventuelle que l'on a
	  // jusqu'a trouver une vritable trame, alors on dpile
	  // l'horodatage de NB_DEBUT_TRAME position

	  // Debut de trame ?
	  if(GPS_TRAME[i+1] == 0x44 && GPS_TRAME[i+2] == 0x12)
	    {
	      // type de message
	      mot.valeur  = GPS_TRAME[i+6];

#ifdef DEBUG_OEM4_TRAME_COMPUTED
	      printf("------> TRAME BINAIRE\n");
	      printf("Etat b4: %d\n",mot.etat.b4);
	      printf("Etat b5: %d\n",mot.etat.b5);
#endif

	      // calcul longueur header
	      HEADER_LENGTH = GPS_TRAME[i+3];

	      // calcul longueur message
	      ((char*)& MESSAGE_LENGTH)[0] = GPS_TRAME[i+8];
	      ((char*)& MESSAGE_LENGTH)[1] = GPS_TRAME[i+9];

	      // Routage de la trame selon son type :
	      //  - binary,
	      //  - ASCII,
	      //  - Abreviated, NMEA
	      //  - Reserved

	      int totLength = MESSAGE_LENGTH + HEADER_LENGTH + 4;

	      // Est-ce que la trame est complete ? si oui on la dtrame  condition
	      // que le buffer contiennent un message de taille MESSAGE_LENGTH + HEADER_LENGTH + 4 (CRC)
	      if( LAST_POSITION - i >= totLength)
		{
		  // printf("\nDestock index = %d\n", i);
		  // infos d'horodatage
		  // printf("------> LAST TIME\n");
		  // printHorodatage(&timeArrival[i]);
		  // printf("------> LAST TIME\n");

		  // Detramage en fonction du type de trame
		  // Detramage binaire
		  if(mot.etat.b5 == 0 && mot.etat.b6 == 0)
		    {
		      // On calcul la distance entre le dbut de la trame et la fin
		      // jusqu'au CRC, cette taille sera ensuite dpile du buffer
		      // pour tre ensuite r-utilise
		      BUFFER_TRAITE = i + totLength;

		     // DEBUG_FUNCTION(HEADER_LENGTH, MESSAGE_LENGTH, ind, BUFFER_TRAITE);

		      // On traite le message recu, la trame commence  ind, le header fait HEADER_LENGTH
		      // octet, le message fait MESSAGE_LENGTH octet, on traite le "NB_TRAME_BUFFEREDime" message
		      if (computeBinaryMessage(GPS_TRAME, i, HEADER_LENGTH, MESSAGE_LENGTH,
					   NB_TRAME_BUFFERED, timeArrival[i]))
					{
						// the message couldn't be read! (probably crc error)
						i++;
						continue;
					}

		      // Incrment du nombre de trame stocke
		      NB_TRAME_BUFFERED++;

		      // On deplace la lecture dans le buffer de la taille de la trame
		      // traite et stocke (on enleve -1 car on commence  zro)
		      i = BUFFER_TRAITE - 1;
		    }
		}//fi
	    }//fi
	}//fi
    }//for

  //printf("\n");

#ifdef DEBUG_DETRAME
  printf("\n\n############################################## TOTAL \n");
  int debutBuff;
  for(debutBuff=0;debutBuff<LAST_POSITION;debutBuff++)
    {
      printf("%0x -",GPS_TRAME[debutBuff]);
    }
  printf("\ntotal buffer = %i\n",debutBuff);
#endif

  // Si des trames ont t traites alors on dpile le buffer
  // des data traits et on remet  jour LAST_POSITION
  if(BUFFER_TRAITE > 0)
    {
      // On depile les informations  partir de BUFFER_TRAITE jusqu'a LAST_POSITION
      // car jusqu'a BUFFER_TRAITE les infos ont t dtrames et places dans une
      // structures de donnes, il n'est plus ncessaire de les conserver dans le buffer,
      // on les dpile en sauvegardant les infos qui restent de BUFFER_TRAITE  LAST_POSITION
      unStackBuffer(BUFFER_TRAITE,  LAST_POSITION);
      unStackArrivalTime(BUFFER_TRAITE,  LAST_POSITION);

      // MAJ de la position du dernier octet dans le buffer et de l'horodatage
      LAST_POSITION = LAST_POSITION - BUFFER_TRAITE;

      // printf("###########################################  NEW POSITION = %d\n", LAST_POSITION);

#ifdef DEBUG 
      // Log des traces
      logTrame();
#endif
      
      // on renvoie toutes les trames taites
      return  NB_TRAME_BUFFERED;
    }
  else
    return -1;
}//fin


/*!
  @brief Cette procdure permet de dpiler les donnes recues du GPS et de les replacer dans
  le buffer remis  zro.

  @param int bufferCompleted : position jusqu' laquelle toutes les donnes ont t traites
  @param int lastPosition : position de fin de donnes dans le buffer
*/
void unStackBuffer(int bufferCompleted,  int lastPosition)
{
  // On nettoie le buffer de stockage temporaire
  memset(GPS_TRAME_COMPLETED, 0, TAILLE_BUFFER);

  int i, ind=0;

  // On se place  la fin des data traites dans le buffer et on rcupre
  // dans le buffer de sauvegarde les data qui ne sont pas encore traites
  for(i= bufferCompleted; i< lastPosition;i++)
    {
      GPS_TRAME_COMPLETED[ind] = GPS_TRAME[i];

      ind++;
    }

  // On nettoie le buffer de stockage
  memset(GPS_TRAME, 0, TAILLE_BUFFER);

  // On recopie dans le buffer de trame les infos qui n'ont pas t encore
  // traites, la distance est donne par (lastPosition - bufferCompleted)
  for(i= 0; i< lastPosition - bufferCompleted;i++)
    {
      GPS_TRAME[i] = GPS_TRAME_COMPLETED[i];
    }
}

/*!
  @brief Cette fonction permet d'envoyer des ordres au GPS dans un format
  ASCII compress, voir doc NOVATEL. Attention certaines commandes demande
  un LAPS de temps important entre deux ordre (reboot du GPS).

 @param int fd: descripteur de fichier sur le GPS (/dev/ttyS0 ...)
 @param char buf[] : commande que l'on veut envoyer au GPS
*/
int writeSerial(int fd, char buf[])
{
  /* for now */
  int state = write(fd, buf, strlen(buf) + 1);

  // erreur d'ecriture renvoie -1 avec errno positionn
  if (state == -1)
    {
      fprintf(stderr, "write error : %s:%d: tty: %s\n", __FILE__, __LINE__, strerror(errno));
      return -1;
    }
  else if (state == 0) // aucun caractre crit sur le port srie, errno non positionn
    {
      perror("write");
      return 0;
    }

   return state;
}

/*!
  @brief : fonction de debug
*/
void DEBUG_FUNCTION( unsigned char HEADER_LENGTH, unsigned short MESSAGE_LENGTH, int ind, int BUFFER_TRAITE)
{
  printf("FORMAT BIN, ind = %d \n", ind);

  int debut, cnt=0;

  printf("############################################## buffer Traite \n");
  printf("HEADER\n");

  for(debut=ind;debut<(ind + HEADER_LENGTH);debut++)
    {
      printf("%0x -",GPS_TRAME[debut]);
      cnt++;
    }

  printf("\ntraite = %i\n",cnt);
  cnt=0;
  printf("\nMESSAGE\n");

  for(debut=ind + HEADER_LENGTH;debut<(ind + HEADER_LENGTH + MESSAGE_LENGTH);debut++)
    {
      printf("%0x -",GPS_TRAME[debut]);
      cnt++;
    }

  printf("\ntraite = %i\n",cnt);
  cnt=0;
  printf("\nCRC\n");

  for(debut=ind + HEADER_LENGTH + MESSAGE_LENGTH;
      debut<(ind + HEADER_LENGTH + MESSAGE_LENGTH + 4);
      debut++)
    {
      printf("%0x -",GPS_TRAME[debut]);

      cnt++;
    }

  printf("\ntraite = %i\n",cnt);

  cnt=0;

  printf("total traite = %i\n", BUFFER_TRAITE);

}

/*!
  @brief : check if the GPS is connected to the serial port
 */
int checkGPS(int fd) {

  printf("check GPS ...\n");

  int nbTrame=0;

  clearBufferSerie(fd);
  
  writeSerial(fd,"UNLOGALL\n");
  sleep(1);
  readSerial(fd,&nbTrame);

  /* if the device returns OK .. suppose that the device is ready */
  if (BUFFER_SERIE[2]=='<' && BUFFER_SERIE[3]=='O' && BUFFER_SERIE[4]=='K')
    return 1;
  else 
    return 0;

}
